This is a replication package for the paper "Prevalence of Botched Code
Integrations" by Ward Muylaert and Coen De Roover submitted to the Mining
Challenge of the 2017 Mining Software Repositories conference.

# Dataset

## Origin of Dataset

We use the [GHTorrent](http://ghtorrent.org/) (version 2016-05-04) and
[TravisTorrent](https://travistorrent.testroots.org/) (version 2016-12-06)
datasets as a basis. Download the MySQL dumps from their respective websites.

## Refining the Dataset

### Projects with 50+ builds of merge commits

The following query fills up a `travistorrent_projects_with_enough_merges`
table.  Projects in this table have 50+ merge commits. A CSV dump of all
projects and their number of merge commits is provided in
`mergesperproject.csv`.

```sql
INSERT INTO `travistorrent_projects_with_enough_merges` (
  gh_project_name,
  amount_of_merges
)
(
  SELECT gh_project_name,
         COUNT(*) AS merges
  FROM (
    SELECT c.sha,dc.gh_project_name
    FROM (
      SELECT DISTINCT git_trigger_commit,gh_project_name
      FROM `travistorrent_6_12_2016`
    ) AS dc
    INNER JOIN commits AS c
    ON c.sha = dc.git_trigger_commit
    INNER JOIN commit_parents AS cp
    ON c.id = cp.commit_id
    GROUP BY c.id
    HAVING COUNT(*) > 1
  ) AS merge_commits
  GROUP BY gh_project_name
  HAVING merges > 49
);
```

### Select projects that follow CI

First, get status data for every project. The result of this query is provided
in `tr_status_per_project.csv`.

```sql
SELECT gh_project_name,
       COUNT(*) AS total,
       SUM(tr_status = 'passed'),
       SUM(tr_status = 'failed'),
       SUM(tr_status = 'errored'),
       SUM(tr_status = 'canceled'),
       SUM(tr_status = 'started')
FROM (
  SELECT tt.*
  FROM `travistorrent_6_12_2016` AS tt
  INNER JOIN `travistorrent_projects_with_enough_merges` AS p
  ON p.gh_project_name = tt.gh_project_name
  GROUP BY git_trigger_commit, tr_status
) AS all_builds
GROUP BY gh_project_name;
```

Next, we use the IQR to remove outliers with too low success rate. These
calculations are done in R.

```r
status_per_project <- read.table('tr_status_per_project.csv', header=TRUE, sep=',')
status_per_project$successrate <- status_per_project$passed / status_per_project$total
q1 <- quantile(status_per_project$successrate)[["25%"]]
lowerbound <- q1 - 1.5*IQR(status_per_project$successrate)
status_per_project$enoughsuccess <- status_per_project$successrate > lowerbound
filtered <- status_per_project[status_per_project$successrate > lowerbound,]
write.table(filtered$gh_project_name, "projects_with_enough_merges_and_success.csv", sep=',', row.names=FALSE, col.names=FALSE)
```

This R script creates a CSV file `projects_with_enough_merges_and_success.csv`
which we also provide. The CSV file contains 555 projects. We import the CSV
file into MySQL as the table
`travistorrent_projects_with_enough_merge_and_success`.

### Projects with builds for the parents of merge commits

*Note: This filtering was originally performed in the results section.* The
reviewers rightfully pointed out that it belongs in the Refining Dataset
section. I did not move things around to fit this narrative in this README.
Instead, the R script used to generate the plots accompanying the dataset is
adjusted. (see section of this README: Characterize the Dataset) It now makes
use of a file that is not created until further down in the section Getting
Numbers of Frequency of Conflicts.

## Expanding the Dataset

### Find merge commits and parents

From the projects we care about, find the merges and save their sha as well as
the sha of both parents in the table `travistorrent_mergeparents`. Uses a dirty
trick to get parent 1 and parent 2. The `parent_id` used is numeric so `MIN()`
will get one parent and `MAX()` will get the other. This approach fails for
commits with three or more parents, but there was only one such case in the
selected projects.

```sql
INSERT INTO `travistorrent_mergeparents` (
    gh_project_name,
    sha,
    parent1_sha,
    parent2_sha
)
(
SELECT m.gh_project_name, m.sha, c1.sha AS parent1, c2.sha AS parent2
FROM (
    -- Find the parent IDs too
    SELECT merges.gh_project_name,
           merges.sha,
           MAX(cp.parent_id) AS parent_1_id,
           MIN(cp.parent_id) AS parent_2_id
    FROM (
        -- Get merge GHTorrent ids
        SELECT c.sha,dc.gh_project_name, c.id
        FROM (
            SELECT DISTINCT tt.git_trigger_commit,tt.gh_project_name
            FROM `travistorrent_6_12_2016` AS tt
            INNER JOIN `travistorrent_projects_with_enough_merge_and_success` AS tts
            ON tt.gh_project_name = tts.gh_project_name
            ) AS dc
        INNER JOIN commits AS c
        ON c.sha = dc.git_trigger_commit
        INNER JOIN commit_parents AS cp
        ON c.id = cp.commit_id
        GROUP BY c.id
        HAVING COUNT(*) > 1
        ) AS merges
    INNER JOIN commit_parents AS cp
    ON merges.id = cp.commit_id
    GROUP BY cp.commit_id
    ) AS m
INNER JOIN commits AS c1
ON c1.id = m.parent_1_id
INNER JOIN commits AS c2
ON c2.id = m.parent_2_id
);
```

### Fight lack of normalization

The lack of normalization makes joining terrible at times. We create an
intermediate table `travistorrent_grouped` containing SHAs and `tr_status`
combinations.

```sql
INSERT INTO travistorrent_grouped (git_trigger_commit, tr_status)
(
    SELECT git_trigger_commit, tr_status
    FROM travistorrent_6_12_2016
    GROUP BY git_trigger_commit, tr_status
)
```

### Status for merge commits and their parents

Finally, we create a table containing the SHA of merge commits and their
parents as well as the status for all three commits.

```sql
INSERT INTO travistorrent_mergeparentsstatus (
    gh_project_name,
    sha,
    parent1_sha,
    parent2_sha,
    sha_status,
    parent1_status,
    parent2_status
)
(
    SELECT mp.gh_project_name,
           mp.sha,
           mp.parent1_sha,
           mp.parent2_sha,
           tt.tr_status,
           tt1.tr_status,
           tt2.tr_status
    FROM `travistorrent_mergeparents` AS mp
    INNER JOIN `travistorrent_grouped` AS tt
    ON mp.sha = tt.git_trigger_commit
    INNER JOIN `travistorrent_grouped` AS tt1
    ON mp.parent1_sha = tt1.git_trigger_commit
    INNER JOIN `travistorrent_grouped` AS tt2
    ON mp.parent2_sha = tt2.git_trigger_commit
    GROUP BY mp.sha, tt.tr_status, tt1.tr_status, tt2.tr_status
);
```

## Characterize the Dataset

Gets the number of commits per project. Result saved in
`commitsperproject.csv`.

```sql
SELECT gh_project_name,
       COUNT(*) AS number_of_commits
FROM (
    SELECT DISTINCT git_trigger_commit, gh_project_name
    FROM travistorrent_6_12_2016
) AS distinct_commits
GROUP BY gh_project_name;
```

Gets the maximum team size in github as recorded by travistorrent. Note that
this varies over time which is why we take the maximum. Result saved in
`max_gh_team_size_per_project.csv`.

```sql
select gh_project_name, MAX(gh_team_size) AS max_team_size
from travistorrent_6_12_2016
group by gh_project_name
order by gh_project_name
```

Gets the amount of branches per project. Result saved in
`number_of_branches_per_project.csv`.

```sql
SELECT gh_project_name, COUNT(git_branch) AS number_of_branches
FROM (
    SELECT gh_project_name,git_branch 
    FROM travistorrent_6_12_2016
    GROUP BY gh_project_name,git_branch 
) branches
GROUP BY gh_project_name
ORDER BY gh_project_name;
```

Gets the number of builds per project. Result saved in
`number_of_builds_per_project.csv`.

```sql
SELECT gh_project_name, COUNT(tr_build_id) AS number_of_builds
FROM (
    SELECT gh_project_name,tr_build_id 
    FROM travistorrent_6_12_2016
    GROUP BY gh_project_name,tr_build_id 
) builds
GROUP BY gh_project_name
ORDER BY gh_project_name;
```

Scatterplots used in the paper are then produced with `projectsinfo.r`.

#  Frequency of Conflicts

## Getting numbers

### Regular commits

```sql
SELECT gh_project_name,
       COUNT(*) AS all_builds,
       SUM(status_now='failed' AND status_before='passed') AS fail_after_pass,
       SUM(status_before='passed') AS total_after_pass
FROM (
    SELECT current.gh_project_name,
           current.git_trigger_commit AS current_commit,
           current.tr_status AS status_now,
           previous.tr_status AS status_before,
           previous.git_trigger_commit AS previous_commmit
    FROM (
        SELECT gh_project_name
        FROM travistorrent_mergeparentsstatus
        GROUP BY gh_project_name
        HAVING COUNT(*) > 49
    ) AS projects
    INNER JOIN `travistorrent_6_12_2016` AS current
    ON projects.gh_project_name = current.gh_project_name
    INNER JOIN `travistorrent_6_12_2016` AS previous
    ON current.tr_prev_build = previous.tr_build_id
    GROUP BY current.git_trigger_commit, current.tr_status, previous.tr_status
) AS commitstatus
GROUP BY gh_project_name;
```

Saved as `failafterpass-regular.csv` (provided).

### Merge commits

```sql
SELECT gh_project_name,
       COUNT(*) AS all_builds,
       SUM(sha_status='failed' AND parent1_status='passed' AND parent2_status='passed') AS fail_after_pass,
       SUM(parent1_status='passed' AND parent2_status='passed') AS total_after_pass
FROM travistorrent_mergeparentsstatus
GROUP BY gh_project_name
HAVING COUNT(*) > 49;
```

Saved as `failafterpass-merge.csv` (provided).

#### Pull requests

```sql
SELECT gh_project_name,
       COUNT(*) AS all_builds,
       SUM(sha_status='failed' AND parent1_status='passed' AND parent2_status='passed' AND gh_is_pr=0) AS fail_after_pass_merge,
       SUM(sha_status='failed' AND parent1_status='passed' AND parent2_status='passed' AND gh_is_pr=1) AS fail_after_pass_pr,
       SUM(parent1_status='passed' AND parent2_status='passed') AS total_after_pass,
       SUM(parent1_status='passed' AND parent2_status='passed' AND gh_is_pr=0) AS total_after_pass_merge,
       SUM(parent1_status='passed' AND parent2_status='passed' AND gh_is_pr=1) AS total_after_pass_pr
FROM travistorrent_mergeparentsstatus AS mps
INNER JOIN (
    SELECT mps.`sha`, tt.gh_is_pr
    FROM travistorrent_mergeparentsstatus AS mps
    INNER JOIN travistorrent_6_12_2016 AS tt
    ON mps.sha = tt.git_trigger_commit
    GROUP BY mps.sha) AS nopr
ON nopr.sha = mps.sha
GROUP BY gh_project_name
HAVING all_builds > 49;
```

Saved as `failafterpass-merge-yesnopr.csv` and in table `travistorrent_breaker_merge_pr`.

## Processing numbers on conflictfrequency

See `conflictfrequency.r`.

# Identify Breaking Commits

We create the table `travistorrent_breakingbuilds` with following fields:

- project
- build id
- breaking commit
- is it a merge?

## Regular Commits

Inserting regular commits is done with

```sql
INSERT INTO travistorrent_breakingbuilds (
    gh_project_name,
    tr_build_id,
    git_trigger_commit,
    is_merge)
(
    SELECT gh_project_name,
           breaker_build_id,
           breaker_commit,
           FALSE
    FROM (
        SELECT current.gh_project_name,
               current.git_trigger_commit AS breaker_commit,
               current.tr_build_id AS breaker_build_id,
               current.tr_status AS status_now,
               previous.tr_status AS status_before,
               previous.git_trigger_commit AS previous_commmit
        FROM (
            SELECT gh_project_name
            FROM travistorrent_mergeparentsstatus
            GROUP BY gh_project_name
            HAVING COUNT(*) > 49
        ) AS projects
        INNER JOIN `travistorrent_6_12_2016` AS current
        ON projects.gh_project_name = current.gh_project_name
        INNER JOIN `travistorrent_6_12_2016` AS previous
        ON current.tr_prev_build = previous.tr_build_id
        GROUP BY current.git_trigger_commit, current.tr_status, previous.tr_status
    ) AS commitstatus
    WHERE status_now = 'failed'
    AND status_before = 'passed'
)
```

## Merge Commits

The merge commits are inserted with

```sql
INSERT INTO travistorrent_breakingbuilds (
    gh_project_name,
    tr_build_id,
    git_trigger_commit,
    is_merge
)
(
    SELECT merges.gh_project_name,
           tt.tr_build_id,
           merges.sha,
           TRUE
    FROM travistorrent_mergeparentsstatus AS merges
    INNER JOIN travistorrent_breaker_merge_pr AS projects
    ON merges.gh_project_name = projects.gh_project_name
    INNER JOIN (
        SELECT *
        FROM travistorrent_6_12_2016
        WHERE tr_status = 'failed'
        GROUP BY tr_build_id
    ) AS tt
    ON tt.git_trigger_commit = merges.sha
    WHERE parent1_status = 'passed'
    AND parent2_status = 'passed'
    AND sha_status = 'failed'
)
```

# Effort to Fix Conflicts

Finding the fixing commit is done with the `findfixer.rb` script.  Results are
saved in `travistorrent_fixer` and provided as `travistorrent_fixer.csv`.  Note
that this also gathers information about what was changed between the breaking
commit and the fixing commit.

R script in `fixer.r` creates the plot for the median LINES metric.

# Where is the Effort Located?

R script in `effort_location.r`.
