require 'mysql2'

def log(text)
  puts "#{Time.now.getutc}: #{text}"
end

##
# Returns a mysql client
def connect
  require 'yaml'
  info = YAML.load_file('database.yml')

  log 'Connecting to database...'

  client = Mysql2::Client.new(
    host: info['host'],
    username: info['username'],
    password: info['password'],
    database: info['database']
  )
  return client
end

##
# Contacts database and fetches all the commits identified
# as being "breakers" (fail after pass).
# Returns the mysql reply that can be iterated over.
# Each row has:
#   - gh_project_name
#   - tr_build_id
#   - git_trigger_commit
#   - is_merge
def fetch_failures(client)
  log 'Fetching failing commits...'

  q = "
    SELECT *
    FROM travistorrent_breakingbuilds
  "

  commits = client.query(q)

  log 'Fetched the commits.'

  return commits
end

def prepare_search_query(client)
  txt = "
    SELECT *
    FROM travistorrent_6_12_2016
    WHERE tr_prev_build = ?
    GROUP BY tr_build_id
    ORDER BY tr_build_id ASC
  "
  return client.prepare(txt)
end

def prepare_insert_query(client)
  txt = "
    INSERT travistorrent_fixer (
      gh_project_name,
      break_tr_build_id,
      fix_tr_build_id,
      nr_builds,
      nr_commits,
      git_diff_src_churn,
      git_diff_test_churn,
      gh_diff_files_added,
      gh_diff_files_deleted,
      gh_diff_files_modified,
      gh_diff_tests_added,
      gh_diff_tests_deleted,
      gh_diff_src_files,
      gh_diff_doc_files,
      gh_diff_other_files,
      is_merge,
      duration
    )
    VALUES (
      ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?
    )
  "
  return client.prepare(txt)
end

def prepare_find_time_query(client)
  txt = "
    SELECT gh_build_started_at
    FROM travistorrent_6_12_2016
    WHERE tr_build_id = ?
  "
  return client.prepare(txt)
end

def get_time(q, build_id)
  res = q.execute(build_id)
  return res.take(1)[0]['gh_build_started_at']
end

# 1. Collect list of failing merges (repo id, branch, build number)
# 2. For each find next one that passes
# 3. Insert breaker id and fixer id in a table

client = connect()
commits = fetch_failures(client)
search_query = prepare_search_query(client)
insert_query = prepare_insert_query(client)
time_query = prepare_find_time_query(client)


ctr = 0
nadactr = 0
commits.each do |commit|
  begin
    if ctr % 1000 == 0
      log "Handled #{ctr} commits. No next found for #{nadactr} commits."
    end
    ctr = ctr + 1
    
    children = []
    parent = commit

    while parent['tr_status'] != 'passed'
      res = search_query.execute(parent['tr_build_id'])
      if res.count == 0
        # Did not find something passing, forget all previous commits too and
        # stop looking.
        nadactr = nadactr + 1
        children = []
        break
      end
      res = res.take(1)[0] # for now only care about the very next one
      children.push(res)
      parent = res
    end

    # Once out the while loop, there are two possibilities:
    # 1. children is empty: at some point while looking for a passing child, we ran out of children
    # 2. children is not empty: passing build found (and in last position), children also holds all the intermediate commits
    if children.empty?
      next # Skip this breaking commit
    else
      # Handle children and save info.
      gh_project_name = commit['gh_project_name']
      break_tr_build_id = commit['tr_build_id']
      fix_tr_build_id = children[-1]['tr_build_id']
      nr_builds = children.length
      nr_commits = children.inject(0) { |sum, entry| sum + entry['git_num_all_built_commits'] }
      git_diff_src_churn = children.inject(0) { |sum, entry| sum + entry['git_diff_src_churn'] }
      git_diff_test_churn = children.inject(0) { |sum, entry| sum + entry['git_diff_test_churn'] }
      gh_diff_files_added = children.inject(0) { |sum, entry| sum + entry['gh_diff_files_added'] }
      gh_diff_files_deleted = children.inject(0) { |sum, entry| sum + entry['gh_diff_files_deleted'] }
      gh_diff_files_modified = children.inject(0) { |sum, entry| sum + entry['gh_diff_files_modified'] }
      gh_diff_tests_added = children.inject(0) { |sum, entry| sum + entry['gh_diff_tests_added'] }
      gh_diff_tests_deleted = children.inject(0) { |sum, entry| sum + entry['gh_diff_tests_deleted'] }
      gh_diff_src_files = children.inject(0) { |sum, entry| sum + entry['gh_diff_src_files'] }
      gh_diff_doc_files = children.inject(0) { |sum, entry| sum + entry['gh_diff_doc_files'] }
      gh_diff_other_files = children.inject(0) { |sum, entry| sum + entry['gh_diff_other_files'] }
      is_merge = commit['is_merge']
      duration = children[-1]['gh_build_started_at'] - get_time(time_query, commit['tr_build_id'])
      insert_query.execute(
        gh_project_name,
        break_tr_build_id,
        fix_tr_build_id,
        nr_builds,
        nr_commits,
        git_diff_src_churn,
        git_diff_test_churn,
        gh_diff_files_added,
        gh_diff_files_deleted,
        gh_diff_files_modified,
        gh_diff_tests_added,
        gh_diff_tests_deleted,
        gh_diff_src_files,
        gh_diff_doc_files,
        gh_diff_other_files,
        is_merge,
        duration
      )
    end
  rescue Mysql2::Error => e
    # Ugly, but Mysql2 does not give different error types...
    # What we are catching is the possibility that the prepared statements is
    # thrown out by MySQL, which is apparently a real possibility.
    # See http://stackoverflow.com/a/33739952/411495
    if e.message == 'Attempt to read a row while there is no result set associated with the statement'
      # Reprepare statements
      search_query = prepare_search_query(client)
      insert_query = prepare_insert_query(client)
      time_query = prepare_find_time_query(client)
      # And do this loop again
      redo
    else
      raise
    end
  end
end

log "Handled #{ctr} commits. No next found for #{nadactr} commits."
